/*
** Module   : integer.java
** Abstract : Progress 4GL compatible integer object
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description----------------------------
** 001 GES 20050517   @21251 Created initial version with full support
**                           for all Progress integer processing including
**                           some functions like integer, minimum and
**                           maximum.
** 002 GES 20050607   @21443 Minor additions for function support.
** 003 GES 20050614   @21489 Added increment and decrement methods.
** 004 GES 20050621   @21533 Added an assign() method to modify this
**                           instance's state based on the result of
**                           an expression of compatible type.
** 005 SVG 20050720   @21767 Cloning support for BaseDataType added.
** 006 GES 20050822   @22192 Added Undoable interface support.
** 007 GES 20050830   @22384 Added error condition processing.
** 008 ECF 20051004   @22970 Implement hashCode() method. Required to be
**                           consistent with BaseDataType's implementation
**                           of equals() method.
** 009 GES 20060123   @24024 Implemented support for decimal and group
**                           separator overrides.
** 010 GES 20060206   @24301 Made compareTo() safe to process any object
**                           type (avoids ClassCastException).
** 011 GES 20060405   @25368 Support any NumberType in lists for min/max.
** 012 GES 20060417   @25555 Added default value generation.
** 013 GES 20060530   @26665 Fixed issue with "hh:mm" formatting where the
**                           hours could be returned with only 1 digit.
** 014 NVS 20060602   @26937 Fixed assign() method for unknown values.
** 015 GES 20060724   @28162 Added a forced assign() abstract method.
** 016 GES 20060807   @28474 Moved to externalizable interface to optimize
**                           network performance.
** 017 ECF 20060812   @28588 Added methods needed for database server-side
**                           function support.
** 018 GES 20060825   @28900 Added context local default format data.
** 019 SIY 20070405   @32825 Part of numbers conversion quirk handling.
** 020 GES 20080911   @39790 Added assign(int) and assign(double)
**                           signatures.
** 021 ECF 20090212   @41299 Implemented a faster equals(). Handles most
**                           common cases without having to invoke the
**                           superclass' implementation. Also added
**                           several assign() variants and optimized
**                           compareTo() for the common case.
** 022 SVL 20090225   @41367 Make class not final.
** 023 SVL 20090824   @43722 "force" parameter should be false by default into
**                           assign(integer).
** 024 SVL 20090908   @43845 Fixed handling of large integers.
** 025 GES 20091102   @44301 Time format string processing was incorrect when
**                           the value is a large enough negative to cross
**                           back over the DST boundary. Removed all J2SE
**                           date/time processing and duplicated the Progress
**                           result directly.
** 026 GES 20110303          Moved maximum/minimum to varargs.
** 027 OM  20121220          Added int64 support.
** 028 CA  20130123          Added c'tor which builds a new instance from a
**                           BaseDataType value, if possible. This should be
**                           used only when constructing an instance using the
**                           value returned by a DYNAMIC-FUNCTION call; 4GL
**                           does some special casting in this case, which is
**                           not implemented yet.
** 029 CS  20130124          Added conversion support for memptr related 
**                           functions task #1635
** 030 GES 20130126          Changed some constructors to handle NumberType.
** 031 GES 20130322          Fixed the BDT c'tor and modified the assign(BDT,boolean) to be the
**                           common location for the _POLY morphing. This is safe because all
**                           current usage would have generated a ClassCastException if that
**                           method was ever used with an invalid type. Now, instead of the CCE,
**                           the code will morph the value as is possible. The morphing code
**                           still needs to be implemented.
** 032 OM  20130328          Added missing runtime implementation. Changed the architecture so 
**                           that this class is derived from int64. All assignments will call 
**                           checkBounds in order to check if the new value fits inside 32 bit 
**                           bounds. Most of the duplicate code common with int64 was removed 
**                           and will be accessible through ingeritance.
**
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * A class that represents a Progress 4GL compatible integer object whose
 * data is mutable.  This class provides the basic wrappering of the 
 * primative data, instantiation, type conversion, <code>unknown value</code>
 * representation and the core comparison logic.
 * <p>
 * All Progress language features related to integer/decimal data
 * types are implemented in the {@link MathOps} class (except comparison
 * logic which is in {@link CompareOps}.  All Progress integer/decimal
 * features are supported including all functionality that can be accessed
 * via operators and built-in Progress functions.
 *
 * @author    GES
 */
public class integer
extends int64
{
   /**
    * Default constructor which creates an instance representing the <code>unknown value</code>.
    */
   public integer()
   {
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a function or method with polymorphic return type into the
    * expected type (i.e. DYNAMIC-FUNCTION()).  In such cases, the 4GL does some
    * automatic type conversion (see {@link #assign(BaseDataType, boolean)}).
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public integer(BaseDataType value)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         assign(value, false);
      }
   }

   /**
    * Constructs an instance after casting the parameter into the internal representation of this 
    * class.  If the value is larger than <code>Integer.MAX_VALUE</code> or smaller than 
    * <code>Integer.MIN_VALUE</code> (do not fit into the 32 bit representation) it
    * cannot be represented and error 15747 will be raised: "Value too large to fit in INTEGER"
    * <p>
    * It is not possible to yield an <code>unknown value</code> from this constructor without
    * calling the method {@link #setUnknown} separate.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(long value)
   {
      super.setValue(value);
   }

   /**
    * Constructs an instance after casting the parameter into the internal
    * representation of this class.  No loss of precision will occur.
    * <p>
    * It is not possible to yield an <code>unknown value</code> from this
    * constructor without calling the method {@link #setUnknown} separate.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(int value)
   {
      super.setValue(value);
   }

   /**
    * Constructs an instance after casting the parameter into the internal
    * representation of this class.  Loss of precision may occur in this
    * conversion (all digits to the right of the decimal point will be
    * truncated).
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(double value)
   {
      super.setValue(new decimal(value));
   }

   /**
    * Constructs an instance after copying the parameter's data into the
    * internal representation of this class.
    * <p>
    * When the parameter is <code>true</code> the resulting integer value
    * will be 1. Otherwise the value is 0.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(boolean value)
   {
      super.setValue(value);
   }

   /**
    * Constructs an instance after casting the parameter into the internal representation of this
    * class. Very large and very small values that do not fit into the 32 bit representation 
    * cannot be represented.
    * <p>
    * If the parameter represents the <code>unknown value</code>, this instance will also 
    * represent the <code>unknown value</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(int64 value)
   {
      assign(value, true);
   }

   /**
    * Constructs an instance after casting the parameter into the internal representation of this
    * class. Very large and very small values that do not fit into the 32 bit representation 
    * cannot be represented.
    * <p>
    * If the parameter represents the <code>unknown value</code>, this instance will also 
    * represent the <code>unknown value</code>.
    * <p>
    * If the <code>output</code> parameter is set to <code>true</code>, it means that the 
    * referenced value is saved, as the passed instance will be passed to a function call, and 
    * the state might need to be updated on function exit. The only use of this constructor is 
    * when wrapping of a larger type (<code>int64</code> or <code>decimal</code>) value is 
    * necessary to fit an <code>integer</code> parameter ONLY in <code>OUTPUT</code> mode.
    * Using this constructor in other parameter-passing modes is an error: in INPUT only mode, 
    * the 1-parameter constructor should be used and in INPUT-OUTPUT mode the value is passed 
    * directly without wrapping.
    *
    * @param   value
    *          The value to be used for this instance (decimal/int64).
    * @param   output
    *          Flag indicating if this instance is passed to as a reference for an output 
    *          parameter. Always <code>true</code>. 
    */
   public integer(NumberType value, boolean output)
   {
      // as this constructor is used ONLY for OUTPUT parameters, this object will be initialized 
      // with unknown (?) in the init() method of the function block so the actual value of the 
      // parameter can safely be ignored.
      // assign(value, true);
      
      // only the presence of this parameter should suffice, output should be always true
      if (output)
      {
         outputParameterReference = value;
      }
   }
   
   /**
    * Constructs an instance after casting the parameter into the internal representation of this
    * class. No loss of precision will occur.
    * <p>
    * If the parameter represents the <code>unknown value</code>, this instance will also
    * represent the <code>unknown value</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(integer value)
   {
      assign(value, true);
   }

   /**
    * Constructs an instance after casting the parameter into the internal representation of this
    * class. Loss of precision may occur in this conversion. All digits to the right of the 
    * decimal point will be truncated. Very large and very small values cannot be represented.
    * <p>
    * If the parameter represents the <code>unknown value</code>, this instance will also 
    * represent the <code>unknown value</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(decimal value)
   {
      assign(value, true);
   }

   /**
    * Constructs an instance after copying the parameter's data into the
    * internal representation of this class.
    * <p>
    * If the parameter's represents the <code>unknown value</code>, this
    * instance will also represent the <code>unknown value</code>.
    * <p>
    * When the parameter is <code>true</code> the resulting integer value
    * will be 1. Otherwise the value is 0.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(logical value)
   {
      super.setValue(value);
   }

   /**
    * Constructs an instance after converting the string representation
    * of a number into an integer.  Due to rounding, a loss of precision
    * may occur in this conversion (all digits to the right of the decimal
    * point will be truncated).
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(String value)
   {
      setValue(value);
   }

   /**
    * Constructs an instance after converting the string representation
    * of a number into an integer.  Due to rounding, a loss of precision
    * may occur in this conversion (all digits to the right of the decimal
    * point will be truncated).
    * <p>
    * If the parameter is <code>unknown value</code> the resulting integer is 
    * the <code>unknown value</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(character value)
   {
      if (value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         setValue(value.getValue());
      }
   }

   /**
    * Constructs an instance after copying the parameter's data into the
    * internal representation of this class.
    * <p>
    * If the parameter's represents the <code>unknown value</code>, this
    * instance will also represent the <code>unknown value</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(date value)
   {
      if (value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         super.setValue(value.longValue());
      }
   }

   /**
    * Constructs an instance after copying the parameter's data into the
    * internal representation of this class.
    * <p>
    * If the parameter is <code>null</code>, this instance will represent
    * the <code>unknown value</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(Number value)
   {
      if (value == null)
      {
         setUnknown();
      }
      else
      {
         super.setValue(value.intValue());
      }
   }

   /**
    * Constructs an instance after copying the parameter's data into the
    * internal representation of this class.
    * <p>
    * If the parameter is <code>null</code>, this instance will represent
    * the <code>unknown value</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(Boolean value)
   {
      if (value == null)
      {
         setUnknown();
      }
      super.setValue(value);
   }

   /**
    * Constructs an instance after copying the parameter's data into the
    * internal representation of this class.
    * <p>
    * If the parameter is <code>null</code>, this instance will represent
    * the <code>unknown value</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public integer(Date value)
   {
      this(new date(value));
   }

   /**
    * Hash code implementation which is consistent with {@link BaseDataType#equals}.
    *
    * @return  Hash code value for this object instance.
    */
   public int hashCode()
   {
      // TODO: should all integer-type objects return same hash for same value ??
      return super.hashCode();
      //int result = 17;
      //if (isUnknown())
      //{
      //   return result;
      //}
      //
      //result = 37 * result + value;
      //
      //return result;
   }

   /**
    * Creates a new instance of the same type that represents the <code>unknown value</code>.
    *
    * @return  An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()
   {
      return new integer();
   }

   /**
    * Creates a new instance of the same type that represents the  default initialized value.
    *
    * @return  An instance that represents the default value.
    */
   public BaseDataType instantiateDefault()
   {
      return new integer(0);
   }

   /**
    * Does the same as standard <code>clone()</code> method but returns an instance of 
    * <code>BaseDataType</code> and doesn't throw the <code>CloneNotSupportedException</code>.
    *
    * @return  A clone of this instance.
    */
   public BaseDataType duplicate()
   {
      return new integer(this);
   }

   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(Undoable value)
   {
      if (value instanceof BaseDataType)
      {
         assign((BaseDataType)value, false);
      }
      else
      {
         throw new RuntimeException("Assignement of " + value.getClass() + 
               " to integer not implemented.");
      }
   }

   /**
    * Replacement for the default object reading method. The latest state is read from the 
    * input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      int value       = in.readInt();
      boolean unknown = in.readBoolean();
      
      if (unknown)
      {
         setUnknown();
      }
      else
      {
         super.setValue(value);
      }
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt((int)getValue());
      out.writeBoolean(isUnknown());
   }

   /**
    * Checks if <code>value</code> can be assigned to this data type. This is important for
    * <code>integer</code> objects where the values must fit into 32 bit storage bounds.
    * 
    * @param   value
    *          The value
    *
    * @return  <code>true</code> if value fits into 32 bit storage bounds.
    *
    * @throws  ErrorConditionException
    *          Thrown if the <code>value</code> do not fit within this data-type.
    */
   @Override
   protected boolean checkBounds(long value)
   throws ErrorConditionException
   {
      if (value > Integer.MAX_VALUE || value < Integer.MIN_VALUE)
      {
         ErrorManager.recordOrThrowError(15747,
                                         "Value " + value + " too large to fit in INTEGER");
         return false;
      }

      return true;
   }
}
